import cv2
import sys
import numpy as np
import logging
import skimage

from abc import ABCMeta
from dataclasses import dataclass, field
from typing import Tuple, Optional, List, Dict, Set
from albumentations import ReplayCompose

from albumentations.core.composition import TransformType  # noqa

from dataspree.platform_sdk.decoder.opencv_decoder import OpencvDecoder

from dataspree.inference.ai.models.base_model import BaseModel, InferenceMode, PreprocessedContent
from dataspree.inference.core.exceptions import (InvalidInferenceProcessing, InvalidInferenceConfiguration,
                                                 InvalidInferenceInput, InferenceImageMissing)
from dataspree.inference.core.item import Item
from dataspree.inference.processing.keypoint_utils import (keypoints_trivial_lookup, keypoints_to_box_corners,
                                                           dlds_to_keypoints)

from dataspree.inference.core.regions_of_interest import copy_all_rois

try:
    from dataspree.inference.processing.augmentations.transform import *  # noqa: F403

    logger = logging.getLogger(__name__)


except Exception as e:
    # Custom transformations could not be included (potentially because optional packages are not installed)
    # This is not a problem if there are no preprocessing augmentations to be performed.
    logger = logging.getLogger(__name__)
    logger.info(f'The preprocessing module could not be imported ({e}). Thus, it is impossible to apply the same '
                'preprocessing steps as performed during training. If this feature is requested later on, '
                'Dataspree Inference will report an ImportError.')


@dataclass
class RegionOfInterest:
    """
    Represents a region of interest within an image.

    Attributes:
        name (Optional[str]): The name or identifier of the patch.

        x_left (int): The x-coordinate of the left boundary of the patch.

        x_right (int): The x-coordinate of the right boundary of the patch.

        y_top (int): The y-coordinate of the top boundary of the patch.

        y_bottom (int): The y-coordinate of the bottom boundary of the patch.

        results (List[Dict]): A list of dictionaries containing preprocessing results associated with the patch.

        keypoints (np.ndarray): Keypoints representing the base image.
            Shape: (3, 2), (center, top, right) x (x, y).

        instances (int): Albumentations instances associated to the keypoints.

        positions (int): Albumentations positions associated to the keypoints.
    """

    name: Optional[str]
    x_left: int
    x_right: int
    y_top: int
    y_bottom: int
    results: List[Dict] = field(default_factory=list)

    def __post_init__(self) -> None:
        """
        Initializes the RegionOfInterest object after it has been constructed.

        Calculates keypoints, instances, and positions associated with the patch.

        Raises:
            AssertionError: If the shape of keypoints does not match the expected shape.
                            This implies an implementation error.
        """
        # Center, top, right
        base_height, base_width = self.y_bottom - self.y_top, self.x_right - self.x_left
        boxes = np.array([base_width / 2.0, base_height / 2.0, base_width, base_height, 0])[None]
        self.keypoints = dlds_to_keypoints(boxes)
        self.instances, self.positions = keypoints_trivial_lookup(len(self.keypoints))
        assert self.keypoints.shape == (3, 2)


@dataclass
class PreprocessedImage(PreprocessedContent):
    """
    Represents a preprocessed image for inference, containing regions of interest and original image shape.

    Attributes:
        inference_batches (np.ndarray): Image data for inference.
            Shape: (number of rois, number of crops, number of tta, network input specific).
            Network input shape is channel-first.

        image_context (Optional[np.ndarray]): Image context used for visualizing heatmaps.

        rois (List[RegionOfInterest]): List of regions of interest within the image.

        original_image_shape (Tuple[int, int]): Shape of the original image (height, width).
    """

    rois: List[RegionOfInterest]
    original_image_shape: Tuple[int, int]

    @property
    def n_channels(self) -> int:
        """ Number of channels in the preprocessed image. """
        return self.inference_batches.shape[3]

    @property
    def input_shape(self) -> Tuple:
        """ Shape of the input data. """
        return self.inference_batches.shape[4:]


class ImageBaseModel(BaseModel[PreprocessedImage], metaclass=ABCMeta):
    """
     A base model for inference tasks on images.

     Attributes:
         inference_mode (InferenceMode): The mode of inference (regions of interest / whole image / ...).

         selected_rois (Optional[List[str]]): The list of selected regions of interest (ROIs).

         preprocessing (Optional[Dict]): Dictionary specifying preprocessing configurations.

         test_time_augmentation_horizontal (bool): Whether to apply horizontal test time augmentation.

         test_time_augmentation_vertical (bool): Whether to apply vertical test time augmentation.

         test_time_augmentation_rotate (bool): Whether to apply rotational test time augmentation.

         show_heatmaps (bool): Whether to compute heatmaps.

         input_shape (Tuple[int, ...]): Input shape of the network model.
     """

    def __init__(self, inference_mode: InferenceMode, selected_rois: Optional[List[str]], preprocessing: Optional[Dict],
                 class_labels: Optional[Dict], model_parameters: Dict,
                 test_time_augmentation_horizontal: bool, test_time_augmentation_vertical,
                 test_time_augmentation_rotate: bool, show_heatmaps: bool,
                 input_shape: Tuple[int, ...]) -> None:
        """
        Initializes the ImageBaseModel.

        Args:
            inference_mode (InferenceMode): The mode of inference.

            selected_rois (Optional[List[str]]): The list of selected regions of interest (ROIs).

            preprocessing (Optional[Dict]): Dictionary specifying preprocessing configurations.

            class_labels (Optional[Dict]): A dictionary mapping model class indices to class labels.

            model_parameters (Dict): A dictionary containing model parameters.

            test_time_augmentation_horizontal (bool): Whether to apply horizontal test time augmentation.

            test_time_augmentation_vertical (bool): Whether to apply vertical test time augmentation.

            test_time_augmentation_rotate (bool): Whether to apply rotational test time augmentation.

            show_heatmaps (bool): Whether to compute heatmaps.

            input_shape (Tuple[int, ...]): Input shape of the network model.
        """
        BaseModel.__init__(self)
        self.inference_mode: InferenceMode = inference_mode
        self.selected_rois: List[str] = [] if selected_rois is None else selected_rois
        self.preprocessing: Optional[Dict] = preprocessing
        self.class_labels: Optional[Dict] = class_labels
        self.model_parameters: Optional[Dict] = model_parameters

        self.test_time_augmentation_horizontal: bool = test_time_augmentation_horizontal
        self.test_time_augmentation_vertical: bool = test_time_augmentation_vertical
        self.test_time_augmentation_rotate: bool = test_time_augmentation_rotate

        self.show_heatmaps: bool = show_heatmaps
        self.input_shape: Tuple[int, ...] = input_shape

        if self.preprocessing and sys.modules.get('dataspree.inference.processing') is None:
            raise InvalidInferenceProcessing('This model is configured to apply preprocessing augmentations.'
                                             'However, the required preprocessing package could not be loaded.')

        self.preprocessing_mesh_crop: Optional[MeshCrop] = None
        self.preprocessing_augmentations: Optional[TransformType] = None
        if self.preprocessing:
            self.preprocessing_augmentations = ReplayCompose._restore_for_replay(self.preprocessing)  # noqa
            mesh_crops = [f for f in self.preprocessing_augmentations.transforms if isinstance(f, MeshCrop)]
            if len(mesh_crops) == 1:
                self.preprocessing_mesh_crop = mesh_crops[0]
            elif len(mesh_crops) > 1:
                raise InvalidInferenceProcessing('Invalid preprocessing configuration: Only one mesh crop is allowed.')

    @property
    def ai_platform_model_id(self) -> Optional[int]:
        try:
            return int(self.model_parameters.get('ai_platform_model_id'))
        except (ValueError, TypeError):
            return None

    @property
    def preprocessing_enabled(self) -> bool:
        return bool(self.preprocessing)

    @property
    def number_of_tta_images(self) -> int:
        """
        Calculates the number of Test Time Augmentation (TTA) images generated from a single source image.

        If only rotate is selected, utilize three rotations. Else, use two operations per selected option.
        This is done to avoid redundant computation.

        See :func:`~dataspree.inference.ai.models.image_base_model.ImageBaseModel.tta_apply`.
        """
        return (2 ** sum((self.test_time_augmentation_vertical, self.test_time_augmentation_rotate,
                          self.test_time_augmentation_horizontal))) if \
            (self.test_time_augmentation_horizontal or self.test_time_augmentation_vertical) else 3

    def preprocess(self, item: Item) -> PreprocessedImage:
        """
        Processes the item and extracts `PreprocessedContent` on which inference can be performed.

        Extract regions of interest (ROIs) from the item and preprocess them for inference.

        This step is performed first in the inference pipeline.

        Args:
            item (Item): The input item to be processed.

        Returns:
            PreprocessedImage: The preprocessed content to be used for inference.

        Raises:
            InvalidInferenceProcessing: If the preprocessing cannot be performed.

            NoInferenceImage: If the input item cannot be processed for inference.

            InvalidInferenceConfiguration: If the inference configuration cannot be applied.

        """
        base_image: np.ndarray = ImageBaseModel.get_inference_base_image(self.preprocessing_enabled, item)
        item['inference_base_image'] = base_image
        og_height, og_width = base_image.shape[:2]

        # Assemble rois
        rois: List[RegionOfInterest]
        if self.inference_mode == InferenceMode.WHOLE_IMAGE:
            rois = [RegionOfInterest(None, 0, og_width, 0, og_height)]

        elif self.inference_mode == InferenceMode.ALL_ROIS or self.inference_mode == InferenceMode.SELECTED_ROIS:
            available_rois = copy_all_rois()
            available_rois.update(item.get('rois') or {})

            # TODO: Don't silently discard regions of interest!
            rois = [RegionOfInterest(roi_name, x_left, x_right, y_top, y_bottom)
                    for roi_name, (x_left, x_right, y_top, y_bottom) in (
                        (k, v) for k, v in available_rois.items() if
                        v is not None and (self.inference_mode == InferenceMode.ALL_ROIS or k in self.selected_rois))
                    if 0 <= x_left < x_right <= og_width and 0 <= y_top < y_bottom <= og_height]
        else:
            rois = []

        # Prepare return image
        images = np.empty(
            (len(rois), len(self.preprocessing_mesh_crop.all_crops) if self.preprocessing_mesh_crop else 1,
             1, *self.input_shape[1:]), dtype=np.float32)

        # Fill return image
        for i_patch, patch in enumerate(rois):
            base_patch = base_image[patch.y_top:patch.y_bottom, patch.x_left:patch.x_right]

            if self.preprocessing:
                try:
                    for i_mesh_crop in range(images.shape[1]):
                        if self.preprocessing_mesh_crop is not None:
                            mc = self.preprocessing_mesh_crop.all_crops[i_mesh_crop]
                            self.preprocessing_mesh_crop.apply_crop(mc)

                        # @dev: the order of the arguments is important!
                        rc = self.preprocessing_augmentations(
                            force_apply=True, keypoints=patch.keypoints.copy(), image=base_patch,
                            instances=patch.instances.copy(), positions=patch.positions.copy())
                        images[i_patch, i_mesh_crop] = np.moveaxis(rc['image'], -1, 0)
                        patch.results.append(rc)

                except KeyError as ke:
                    raise InvalidInferenceProcessing(
                        f'Could not replay training augmentations: {ke}. The version of the preprocessing library that '
                        f'was used to augment the image might ship an augmentation ({ke}), that was not implemented '
                        f'in the currently installed version of the preprocessing software Please update the '
                        f'preprocessing software if that is the case.')

            else:
                names: Optional[Tuple[str, ...]] = base_patch.dtype.names
                if names is not None:
                    dtypes: Set[np.dtype] = {base_patch[n].dtype for n in names}
                    if len(dtypes) == 1:
                        base_patch = base_patch.view((next(iter(dtypes)), len(names)))

                base_patch = base_patch.astype(np.float32, copy=False)
                base_patch = cv2.resize(base_patch, (self.input_shape[3], self.input_shape[2]))

                if len(base_patch.shape) < 3:
                    base_patch = base_patch[..., True]

                if self.input_shape[1] == 1 and self.input_shape[1] != base_patch.shape[2]:
                    base_patch = base_patch.mean(2)[..., True]

                elif self.input_shape[1] > 1 >= base_patch.shape[2]:
                    base_patch = np.dstack([base_patch] * self.input_shape[1])

                # transpose from HWC to CHW
                base_patch = base_patch.transpose((2, 0, 1))

                # normalize: base_patch \in [-1, 1]
                base_patch = (2.0 / 255.0) * base_patch - 1

                images[i_patch] = base_patch

        # Generate heat map base image?
        heat_map_base_image = None
        if self.show_heatmaps:
            is_structured_array: bool = base_image.dtype.names is not None

            if not is_structured_array:
                heat_map_base_image = base_image.astype(np.uint8, copy=False)

            else:
                channel_names: List[str]
                if self.preprocessing_augmentations is None:
                    channel_names = list(base_image.dtype.names)
                else:
                    ca_channels: List[List[Dict]] = [f.channels for f in self.preprocessing_augmentations.transforms
                                                     if isinstance(f, ChannelAssembler)]
                    assert len(ca_channels) == 1

                    # Determine channel names
                    channel_names: List[str] = []
                    for ch in ca_channels[0]:
                        selection = ch['channel']['channel_type']['selection']
                        name = ch['channel']['channel_type']['name'] if selection == 'custom' else selection
                        channel_names.append(name)

                if not len(channel_names):
                    raise InvalidInferenceConfiguration('You need to select at least one channel')

                if len(channel_names) > 3:
                    channel_names = channel_names[:3]
                elif len(channel_names) < 3:
                    channel_names = [channel_names[0]]

                heat_map_base_image = np.empty((*base_image.shape, len(channel_names)), dtype=np.uint8)
                for i_channel, channel_name in enumerate(channel_names):
                    c_array = np.empty_like(heat_map_base_image[..., i_channel])
                    cv2.normalize(ChannelAssembler.retrieve_channel(channel_names[0], base_image),
                                  c_array, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
                    heat_map_base_image[..., i_channel] = c_array

            if len(heat_map_base_image.shape) == 3:
                heat_map_base_image = np.mean(heat_map_base_image, axis=-1, dtype=np.uint8)

            if len(heat_map_base_image.shape) != 2:
                raise InvalidInferenceInput(f'Error: The heat map base image has a wrong shape '
                                            f'{heat_map_base_image.shape}.')

        return PreprocessedImage(inference_batches=images, rois=rois, image_context=heat_map_base_image,
                                 original_image_shape=(og_height, og_width))

    def tta_apply(self, processed_input: PreprocessedImage) -> PreprocessedImage:
        """
        Apply test time augmentations (TTA) to the preprocessed input. This function copies the preprocessed
        input and applies different augmentations, which can later be reversed and merged into a single output map.

        This step is performed on the output of `preprocess`.

        Args:
            processed_input (numpy.ndarray): The input data after preprocessing.
                Shape: (number of rois, number of crops, 1, ...)

        Returns:
            numpy.ndarray: The augmented input data.
                Shape: (number of rois, number of crops, number of tta, ...)
        """

        if not any((self.test_time_augmentation_rotate, self.test_time_augmentation_horizontal,
                    self.test_time_augmentation_vertical)):
            return processed_input

        num_rois: int = len(processed_input.rois)
        assert num_rois == processed_input.inference_batches.shape[0]
        num_crops: int = processed_input.inference_batches.shape[1]

        assert processed_input.inference_batches.shape[2] == 1
        output_image = np.empty((num_rois, num_crops, self.number_of_tta_images,
                                 *processed_input.inference_batches.shape[3:]), processed_input.inference_batches.dtype)

        for i_patch in range(num_rois):
            for i_crop in range(num_crops):
                index: int = 0

                og_image = processed_input.inference_batches[i_patch, i_crop]
                output_per_patch = output_image[i_patch, i_crop]

                output_per_patch[index] = og_image
                index += 1

                if self.test_time_augmentation_horizontal:  # noqa Duplicates
                    for i in range(index):
                        output_per_patch[index] = np.flip(output_per_patch[i], axis=-1)
                        index += 1

                if self.test_time_augmentation_vertical:
                    for i in range(index):
                        output_per_patch[index] = np.flip(output_per_patch[i], axis=-2)
                        index += 1

                if self.test_time_augmentation_rotate:

                    for i in range(index):
                        output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                            output_per_patch[i][np.newaxis], cv2.ROTATE_90_CLOCKWISE)
                        index += 1

                        if not self.test_time_augmentation_horizontal and not self.test_time_augmentation_vertical:
                            output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                                output_per_patch[i][np.newaxis], cv2.ROTATE_180)
                            index += 1

                            output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                                output_per_patch[i][np.newaxis], cv2.ROTATE_90_COUNTERCLOCKWISE)

                            index += 1

                assert index == output_image.shape[2]

        return PreprocessedImage(inference_batches=output_image, rois=processed_input.rois,
                                 image_context=processed_input.image_context,
                                 original_image_shape=processed_input.original_image_shape)

    def tta_revert(self, model_output: List[np.ndarray]) -> List[np.ndarray]:
        """
        Reverts the output of the model by applying inverse operations for Test Time Augmentation (TTA) input data.

        Args:
            model_output (List[np.ndarray]): A list of NumPy arrays containing the model's output.

        Returns:
            List[np.ndarray]: A list of NumPy arrays with the TTA inversions applied, restoring the output's link
                              to the original input image.
        """
        num_rois: int = model_output[0].shape[0]
        num_crops: int = model_output[0].shape[1]
        num_tta: int = model_output[0].shape[2]

        out_maps: List[np.ndarray] = []
        for out_map in model_output:
            for i_patch in range(num_rois):
                for i_crop in range(num_crops):
                    output_per_patch = out_map[i_patch, i_crop]

                    index: int = 1
                    if self.test_time_augmentation_horizontal:  # noqa Duplicates
                        for _ in range(index):
                            output_per_patch[index] = np.flip(output_per_patch[index], axis=-1)
                            index += 1

                    if self.test_time_augmentation_vertical:
                        for _ in range(index):
                            output_per_patch[index] = np.flip(output_per_patch[index], axis=-2)
                            index += 1

                    if self.test_time_augmentation_rotate:

                        for _ in range(index):
                            output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                                output_per_patch[index][np.newaxis], cv2.ROTATE_90_COUNTERCLOCKWISE)
                            index += 1

                            if not self.test_time_augmentation_horizontal and not self.test_time_augmentation_vertical:
                                output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                                    output_per_patch[index][np.newaxis], cv2.ROTATE_180)
                                index += 1

                                output_per_patch[index] = ImageBaseModel.rotate_np_array_size_preserving(
                                    output_per_patch[index][np.newaxis], cv2.ROTATE_90_CLOCKWISE)
                                index += 1
                    assert index == num_tta

            out_maps.append(out_map)

        return out_maps

    @staticmethod
    def compute_scaling(preprocessing_enabled: bool, prediction_input: PreprocessedImage) -> np.ndarray:
        """
        Computes the scaling and offset needed to transform patches from the input image to the model's input space.

        Args:
            preprocessing_enabled (bool): Indicates whether preprocessing is enabled.
            prediction_input (PreprocessedImage): The input containing the preprocessed image and its associated data.

        Returns:
            np.ndarray: An array with shape (n_roi, n_crop, 4) containing the scaling and offset information.
                        Each entry contains (x_offset, x_scale, y_offset, y_scale), where:
                        - x_offset, y_offset are in pixels in the input image.
                        - x_scale, y_scale are the scaling factors to apply to heatmaps/detections/...
                          to get the correct size.

        Raises:
            AssertionError: If the positions in `crop_results` do not match the expected values.
                            This implies an implementation error prior to calling this function.

        Notes:
            - The scaling factors are calculated differently based on whether preprocessing is enabled.
            - When preprocessing is enabled, the method accounts for transformations applied during preprocessing,
              such as padding or cropping, to correctly recover the heatmap's position within the patch.
        """
        patch_scaling: np.ndarray = np.empty((prediction_input.n_roi, prediction_input.n_crop, 4), float)

        for i_patch, patch in enumerate(prediction_input.rois):

            patch_offset = np.array([patch.x_left, patch.y_top])
            patch_input_size = np.array((patch.x_right - patch.x_left, patch.y_bottom - patch.y_top))

            if not preprocessing_enabled:
                preprocessing_scaling = patch_input_size / prediction_input.input_shape[::-1]
                input_scaling = preprocessing_scaling

                preprocessing_offset = np.zeros(2) * preprocessing_scaling
                input_offset = preprocessing_offset + patch_offset

                patch_scaling[i_patch, 0] = (input_offset[0], input_scaling[0], input_offset[1], input_scaling[1])

            else:
                # Individual rois have different original sizes.
                # When rectifying the rois to the model size we have different options (such as padding),
                # that cause the resized heatmap to be not exactly at the position of the rois (but rather
                # inside the patch somewhere). Incorporate this here by passing keypoints to the
                # preprocessing operation (0, 0, roi_width, roi_height) and utilizing the transformed
                # positions to correctly recover, among others, the heatmap.

                # TODO: Test that this works also without mesh crop enabled (but with preprocessing).
                for i_crop, crop_results in enumerate(patch.results):
                    assert crop_results['positions'] == ['center', 'top', 'right']

                    # Coordinates of the input image that were transformed by the applied preprocessing operations.
                    transformed_patch_corners = keypoints_to_box_corners(np.array(crop_results['keypoints'])[None])[0]
                    assert transformed_patch_corners.shape == (4, 2)

                    # The original image diagonal divided by the transformed image diagonal
                    preprocessing_scaling: np.ndarray = (
                            patch_input_size / (transformed_patch_corners[2] - transformed_patch_corners[0]))
                    assert preprocessing_scaling.shape == (2,)
                    input_scaling = preprocessing_scaling

                    # The original top left corner is to be scaled by the network distortion and
                    # the preprocessing distortion.

                    # Transformed top-left coordinate of the patch.
                    preprocessing_offset = -transformed_patch_corners[0] * preprocessing_scaling
                    assert preprocessing_offset.shape == (2,)

                    input_offset = preprocessing_offset + patch_offset  # network_output_scaling

                    patch_scaling[i_patch, i_crop] = (input_offset[0], input_scaling[0],
                                                      input_offset[1], input_scaling[1])

        return patch_scaling

    @staticmethod
    def merge_class_activation(prediction_input: PreprocessedImage, class_activation_maps: np.ndarray,
                               total_scaling: np.ndarray) -> Tuple[np.ndarray, List[np.ndarray]]:
        """
        Transforms the class activation maps into the original image space and merges them for regions
        of interest (ROIs) and the entire image.

        Args:
            prediction_input (PreprocessedImage): The input containing the preprocessed image and its associated data.

            class_activation_maps (np.ndarray): The class activation maps.
                                                Shape (n_roi, n_crop, 1, n_classes, height, width).

            total_scaling (np.ndarray): The scaling factors for each ROI and crop
                                        Shape (n_roi, n_crop, 4).

        Returns:
            Tuple[np.ndarray, List[np.ndarray]]:
                - A merged class activation map for the entire image in the original image coordinates.
                - A list of merged class activation maps for each ROI in their respective coordinates.

        Raises:
            AssertionError: If the shape of `class_activation_maps` or `total_scaling` does not match the expected
                            dimensions.
                            This implies an implementation error prior to calling this function.

        Notes:
            - The method assumes that test-time augmentations (TTAs) have already been merged for computing heatmaps.
            - The scaling factors in `total_scaling` should correspond to the coordinates and sizes of the crops within
              the original image space.
        """
        assert prediction_input.n_roi == len(prediction_input.rois)

        assert class_activation_maps.shape[:2] == (prediction_input.n_roi, prediction_input.n_crop)
        assert class_activation_maps.shape[2] == 1, 'TTAs have to be merged for computing heatmaps.'
        assert len(class_activation_maps.shape) == 6, 'n roi, n crop, n tta = 1, n classes, height, width.'

        assert total_scaling.shape == (prediction_input.n_roi, prediction_input.n_crop, 4), \
            'n roi, n crop, (offset x, scale x, offset y, scale y ).'

        n_classes: int = class_activation_maps.shape[3]

        image: Optional[np.ndarray] = prediction_input.image_context
        assert len(image.shape) == 2
        height, width = image.shape

        # This is the merged class activation map for the entire image in the coordinates of the original image.
        merged_image_class_activation = np.zeros((n_classes, height, width), dtype=np.float32)

        # This is the merged class activation map for the separate rois, in the coordinates of the separate rois.
        # Rois differ in sizes.
        merged_roi_class_activations: List[np.ndarray] = []

        for i_roi, roi in enumerate(prediction_input.rois):
            roi_w, roi_h = roi.x_right - roi.x_left, roi.y_bottom - roi.y_top

            merged_roi_class_activation = np.zeros((n_classes, roi_h, roi_w), dtype=np.float32)
            merged_roi_class_activations.append(merged_roi_class_activation)

            for i_crop in range(prediction_input.n_crop):
                class_activation_map = class_activation_maps[i_roi, i_crop, 0]

                # Coordinates of the mesh crop with respect to the original image.
                mc_left, mc_top = map(int, total_scaling[i_roi, i_crop][::2])
                mc_h, mc_w = map(int,
                                 np.array(class_activation_maps.shape[-2:]) * total_scaling[i_roi, i_crop][1::2][::-1])

                resized_class_activation = skimage.transform.resize(class_activation_map,
                                                                    (n_classes, mc_h, mc_w))

                cai = merged_image_class_activation[:, mc_top:mc_top + mc_h, mc_left:mc_left + mc_w]
                np.maximum(cai, resized_class_activation, out=cai)

                cap = merged_roi_class_activation[
                      :, mc_top - roi.y_top:mc_top - roi.y_top + mc_h,
                      mc_left - roi.x_left:mc_left - roi.x_left + mc_w]
                np.maximum(cap, resized_class_activation, out=cap)

        return merged_image_class_activation, merged_roi_class_activations

    @staticmethod
    def class_activation_heatmap(class_activation_map: np.ndarray, image_context: np.ndarray,
                                 use_hsv: bool = False) -> np.ndarray:
        """
        Generates a class activation heatmap by combining a class activation map with an image context.

        Args:
            class_activation_map (np.ndarray): Class activation map. Shape (..., n_classes, height, width).
            image_context (np.ndarray): Image context Shape (..., height, width).
            use_hsv (bool): Whether to use the class activation map for hue and saturation.

        Returns:
            np.ndarray: Heatmap with shape (..., n_classes, height, width, 3).

        Raises:
            AssertionError: If `class_activation_map` is not 3-dimensional or if the shape of `image_context`
                            does not match the shape of the spatial dimensions of `class_activation_map`.
                            This implies an implementation error prior to calling this function.

        Notes:
            - `image_context` can be preprocessed_image.image_context.
        """
        assert len(class_activation_map.shape) >= 3
        assert image_context.shape[:-2] == class_activation_map.shape[:-3]
        assert image_context.shape[-2:] == class_activation_map.shape[-2:]

        heatmap = np.zeros((*class_activation_map.shape, 3), np.uint8)

        if use_hsv:
            heatmap[..., 0] = 255 - class_activation_map * (255 - 180)

            # Set the color to 0 if outside a patch.
            heatmap[..., 1] = np.clip(class_activation_map / class_activation_map.max() * 255 * 10, 0, 255)

            heatmap[..., 2] = (image_context[..., np.newaxis, :, :] * 0.7 + 0.3 * 255)[np.newaxis]

            flat_view = heatmap.view()
            flat_view.shape = (np.prod(heatmap.shape[:-2]), *heatmap.shape[-2:])
            cv2.cvtColor(flat_view, cv2.COLOR_HSV2RGB, flat_view)

        else:
            # TODO: I think I would prefer not showing any "heat" in locations without heatmap.
            #       We can either use the value 0.0 as an indicator for this, as I would expect the output map to
            #       rarely assume 0.0 exactly.

            hm = (class_activation_map * 255).astype(np.uint8)
            hm_flat = hm.view()
            hm_flat.shape = (np.prod(hm.shape[:-1]), *hm.shape[-1:])

            out_flat = heatmap.view()
            out_flat.shape = (out_flat.shape[0] * out_flat.shape[1], *out_flat.shape[2:])

            cv2.applyColorMap(hm_flat, cv2.COLORMAP_JET, dst=out_flat)
            cv2.cvtColor(out_flat, cv2.COLOR_BGR2RGB, dst=out_flat)

            heatmap_intensity = 0.5
            heatmap = (heatmap * heatmap_intensity +
                       image_context[..., np.newaxis, :, :, np.newaxis] * (1 - heatmap_intensity)).astype(
                np.uint8)

        return heatmap

    @staticmethod
    def get_inference_base_image(preprocessing_enabled: bool, item: Item) -> np.ndarray:
        """
        Extracts the image used for inference. If preprocessing is enabled, a structured array is returned.
        If no structured array is found, it is automatically created from the image with the convention from
        #OpencvDecoder.create_structured_array.

        The numpy array found under the key `image` must either be None or no structured array.
        The numpy array found under the key `image_structured` has to be None or a structured array.

        Args:
            preprocessing_enabled (bool): Whether preprocessing is enabled.

            item (Item): The item from which to extract images.

        Returns:
            np.ndarray: The image to be preprocessed.

        Raises:
            InferenceImageMissing: If no suitable image is found.
        """
        image_structured: Optional[np.ndarray] = item.get('image_structured')
        image: Optional[np.ndarray] = item.get('image')

        if not preprocessing_enabled:
            if image is None:
                raise InferenceImageMissing

            if not isinstance(image, np.ndarray):
                raise InferenceImageMissing

            if image.dtype.names is not None:
                raise InferenceImageMissing

            return image

        if image_structured is None and image is None:
            raise InferenceImageMissing

        if image_structured is None:
            if not isinstance(image, np.ndarray):
                raise InferenceImageMissing

            image_structured = OpencvDecoder.create_sturctured_array(image)

        if not isinstance(image_structured, np.ndarray):
            raise InferenceImageMissing

        return image_structured

    @staticmethod
    def efficient_rotate_channel_first(array: np.ndarray,
                                       rotation: int = cv2.ROTATE_90_CLOCKWISE) -> np.ndarray:
        """
        Rotates a 2D or 3D array representing an image by 90, 180, or 270 degrees efficiently.

        Args:
            array (np.ndarray): The input array representing the image to rotate.
                Shape: (number of batches x C x H x W)

            rotation (int, optional): The rotation flag indicating the direction and angle of rotation.
                Defaults to cv2.ROTATE_90_CLOCKWISE, representing a 90-degree clockwise rotation.

        Returns:
            np.ndarray: The rotated image array.

        Raises
        """
        if len(array.shape) != 4:
            raise ValueError('Expecting three four dimensional input array (number of batches x H x W x C) '
                             'or three dimensional input array (number of batches x H x W).')

        if rotation == cv2.ROTATE_90_COUNTERCLOCKWISE:
            return array[:, :, :, ::-1].swapaxes(2, 3)

        elif rotation == cv2.ROTATE_90_CLOCKWISE:
            return array[:, :, ::-1, :].swapaxes(2, 3)

        elif rotation == cv2.ROTATE_180:
            return array[:, :, ::-1, ::-1]

        raise ValueError(f'The rotation {rotation} is not implemented.')

    @staticmethod
    def rotate_np_array_size_preserving(image: np.ndarray, rotation: int) -> np.ndarray:
        """
        Rotates a NumPy array representing an image while preserving its size.

        Args:
            image (np.ndarray): The input image array with dimensions (BxCxWxH), where C represents channels.

            rotation (int, optional): The rotation flag indicating the direction and angle of rotation.
                Defaults to cv2.ROTATE_90_CLOCKWISE, representing a 90-degree clockwise rotation.

        Returns:
            np.ndarray: The rotated image array.

        Notes:
            The function internally swaps the axes of the input array to convert from channel-first (CxWxH)
            to channel-last (WxHxC) format before performing rotation. This ensures compatibility with
            the rotation function used.

        See Also:
            `rotate_channel_first`: The function used internally to efficiently rotate the image.

        Examples:
            # Rotate an RGB image by 90 degrees clockwise
            rotated_image = rotate_np_array_size_preserving(image_rgb, cv2.ROTATE_90_CLOCKWISE)
        """
        # Rotate
        img_rot = ImageBaseModel.efficient_rotate_channel_first(image, rotation)

        # Resize, so that the output array has the same size as the input array
        if image.shape[-1] == image.shape[-2]:
            result = img_rot

        else:
            result = img_rot.view()
            result.shape = np.prod(img_rot.shape[:2]), *img_rot.shape[2:]
            result = cv2.resize(result.T, image.shape[2:]).T
            result.shape = image.shape

        return result
