import logging

from onnx.checker import ValidationError
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Any, Dict, Optional, List

logger = logging.getLogger(__name__)


class ConfigStore(ABC):

    @abstractmethod
    def load_data(self) -> None:
        """
        Load the data from the persistent storage.
        """
        ...

    @abstractmethod
    def save_data(self) -> None:
        """
        Store the data to the persistent storage we read it from.
        """
        ...

    @abstractmethod
    def empty(self) -> bool:
        """
        :return: if the configuration is completely empty.
        """
        ...

    @abstractmethod
    def save_raw_and_reload(self, raw_data: Any) -> None:
        """
        :param raw_data:
        """
        ...

    @abstractmethod
    def get(self, name: str, default: Any = None):
        ...

    @abstractmethod
    def get_data(self) -> Dict:
        """
        :return: copy of the data, changes have no side effect.
        """
        ...

    @abstractmethod
    def put(self, name: str, data: Any) -> None:
        ...

    @abstractmethod
    def put_data(self, data: Any) -> None:
        ...


@dataclass
class RoutingEntry:
    source: str
    sink: str
    filter: Optional[str]


def routing_configuration(cfg: List[Dict[str, Optional[str]]],
                          source: Optional[str], sink: Optional[str]) -> List[RoutingEntry]:
    # TODO: This is terribly inefficient. We should define a schema for the configuration file and create a lookup
    #       table if nothing changed (and provide functions for incremental updates).
    routing_info: List[RoutingEntry] = []
    for rc in cfg:
        try:
            if source is not None and rc['source'] != source:
                continue

            if not isinstance(rc['source'], str):
                raise ValueError('source must be a string')

            if sink is not None and rc['sink'] != sink:
                continue

            if not isinstance(rc['sink'], str):
                raise ValueError('sink must be a string')

            routing_info.append(RoutingEntry(rc['source'], rc['sink'], rc.get('filter')))
        except (ValueError, KeyError):
            logger.warning(f'Configuration entry invalid: {rc}')

    return routing_info
