import copy
from typing import Any, Dict, List, Union, Optional
from itertools import chain
from functools import partial

import yaml
import os
import logging
import traceback

from dataspree.inference.core.config.config_store import ConfigStore
from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.exceptions import DataspreeInferenceException
from dataspree.inference.core.item import Item

logger = logging.getLogger(__name__)


class LinkConfigStore(ConfigStore):

    def __init__(self, root_path: str) -> None:
        # Path to the configuration file.
        self.__root_path: str = os.path.normpath(root_path)

        # The current configuration data, which is not necessarily in-sync with the disk.
        self.__data: Dict = {}

        # Main configuration file abstraction
        self.__root_config: Optional[LocalStorage] = None

        # Loaded configuration files
        self.__loaded_nodes: Dict[str, LocalStorageFile] = {}

    def clear(self):
        self.__data = {}

        root_storage = self.__loaded_nodes[self.__root_path] = LocalStorageFile(self.__root_path, {})

        root_storage_ref = LocalStorageFileReference(root_storage, self.__root_path, False)
        self.__root_config = LocalStorage([], root_storage_ref)
        self.save_data()

    def write_flat_to(self, out_path):

        with open(out_path, 'w') as to:
            yaml.safe_dump(self.__data, to)

    def load_data(self) -> None:
        """
        Parse configuration files, create internal structure that can be accessed through `get_data`.
        """
        loaded_nodes = self.__loaded_nodes = {}
        configuration_file_name = self.__root_config

        class Loader(yaml.SafeLoader):
            """
            Yaml loader that is able to interpret special include directive  !!include.
            """

            def __init__(self, stream):
                yaml.SafeLoader.__init__(self, stream)
                self.__configuration_file_name: str = stream.name
                self.__parent_directory: str = os.path.split(self.__configuration_file_name)[0]

            def include(self, node, as_dict: bool = False):
                include_argument: str = self.construct_scalar(node)
                relative_config_file_path: str = include_argument
                absolute_config_file_path: str = os.path.normpath(os.path.join(self.__parent_directory,
                                                                               relative_config_file_path))

                if os.path.isdir(absolute_config_file_path):
                    absolute_config_file_path = os.path.join(absolute_config_file_path, 'config.yml')
                absolute_config_file_path = os.path.normpath(absolute_config_file_path)

                if absolute_config_file_path not in loaded_nodes:
                    # Prevents circular include problems; Whenever we re-encounter references to the same file,
                    # we return None.
                    # We could also think about returning a proxy to the file that is loaded, in case the circular
                    # import is not problematic (because we are only considering an area that is not circular).
                    loaded_nodes[absolute_config_file_path] = None

                    # Attempt to load, if impossible backup the file and replace with an empty file.
                    content: Dict = {}
                    try:
                        with open(absolute_config_file_path, 'r') as fo:
                            content = yaml.load(fo, Loader)

                    except Exception as e:
                        if not isinstance(e, FileNotFoundError):
                            logger.error(f'Unable to load data from {absolute_config_file_path}. '
                                         f'Override with default. Exception: {e} {loaded_nodes}')

                            # Backup corrupted configuration file.
                            file, extension = os.path.splitext(absolute_config_file_path)
                            os.replace(absolute_config_file_path, f'{file}-backup{extension}')

                        # Reset Configuration and clear configuration file.
                        with open(absolute_config_file_path, 'w') as fo:
                            fo.write(yaml.dump(content))

                    finally:
                        loaded_nodes[absolute_config_file_path] = LocalStorageFile(absolute_config_file_path, content)

                if loaded_nodes[absolute_config_file_path] is None:
                    logger.warning(
                        f'Circular import of file {absolute_config_file_path} detected. Import becomes circular '
                        f'in {configuration_file_name}.')
                    return None

                else:
                    return LocalStorageFileReference(loaded_nodes[absolute_config_file_path],
                                                     relative_config_file_path,
                                                     as_dict=as_dict)

        Loader.add_constructor('!include', Loader.include)
        Loader.add_constructor('!extend', partial(Loader.include, as_dict=True))

        try:
            with open(self.__root_path, 'r') as f:

                root_storage = self.__loaded_nodes[self.__root_path] = LocalStorageFile(
                    self.__root_path, yaml.load(f, Loader))

                root_storage_ref = LocalStorageFileReference(root_storage, self.__root_path, False)
                self.__root_config = LocalStorage([], root_storage_ref)

                self.__data = self.__root_config.get_data()

        except FileNotFoundError:
            self.clear()

        except Exception as e:
            logger.error(f'Could not load data from {self.__root_path}. '
                         f'Override with default. Exception: {e}')

            if is_application_debug():
                traceback.print_exc()

            # Backup corrupted configuration file.
            file, extension = os.path.splitext(self.__root_path)
            os.replace(self.__root_path, f'{file}-backup{extension}')

            # Reset Configuration and clear configuration file.
            self.clear()

    def save_data(self) -> None:

        class Writer(yaml.SafeDumper):
            """
            Yaml writer that is able to write out special include directive !!include.
            """

            def __init__(self, *args, **kwargs) -> None:
                yaml.SafeDumper.__init__(self, *args, **kwargs)

                def to_yaml(dumper, data: LocalStorageFileReference):
                    if data.as_dict:
                        return dumper.represent_scalar('!extend', data.reference)
                    return dumper.represent_scalar('!include', data.reference)

                self.add_multi_representer(LocalStorageFileReference, to_yaml)

        original_data = copy.deepcopy(self.__data)

        if self.__root_config is None:
            raise DataspreeInferenceException('Attempting to write configuration file which has not been read.')

        try:
            # Save data to disk
            self.__root_config.save_data(copy.deepcopy(original_data))

            received_data = self.__root_config.get_data()

            # Sanity check: When regenerated, does the data still match the input data?
            if received_data != original_data:
                raise Exception('Data mismatch.')

            # Reload from disk
            for path, node in self.__loaded_nodes.items():
                with open(path, 'w') as to:
                    yaml.dump(node.config_file_content, to, Writer)

            # Sanity check: does the loaded data still match the input data?
            self.load_data()
            new_data = self.get_data()
            if new_data != original_data:
                raise Exception('Data mismatch after reloading from disk.')

        except Exception as e:

            # The loaded data does not match the input data. This can occur if there's a bug, or for values that are
            # not equal to themselves (e.g., NaN)
            abs_path = self.__root_config.local_storage_file.lsf.config_file_path

            # Create a backup file
            file, extension = os.path.splitext(abs_path)
            os.replace(abs_path, f'{file}-backup{extension}')

            # Flatten configuration into a single configuration file
            self.__loaded_nodes = {abs_path: self.__root_config.local_storage_file.lsf}
            self.__root_config.flatten(original_data)
            self.__data = self.__root_config.get_data()

            new_data = self.get_data()
            if is_application_debug():
                traceback.print_exc()
            logger.error(f'An error occurred while attempting to write configuration file. '
                         f'Flattened match: {original_data == new_data}: {e}')

            # Write the flat configuration file.
            for path, node in self.__loaded_nodes.items():
                with open(path, 'w') as to:
                    yaml.dump(node.config_file_content, to, Writer)

    def save_raw_and_reload(self, raw_data: Any) -> None:
        """
        Override the configuration links with one single configuration file.
        """
        with open(self.__root_path, 'w') as f:
            f.write(raw_data)

        self.load_data()

    def empty(self) -> bool:
        return len(self.__data) == 0

    def get(self, name: str, default: Any = None):
        return self.__data.get(name, default)

    def get_data(self) -> Dict:
        return copy.deepcopy(self.__data)

    def put(self, name: str, data: Any) -> None:
        self.__data[name] = data

    def put_data(self, data: Any) -> None:
        self.__data = data

    def str(self):
        return self.__root_config.str()


# @dataclass
class LocalStorageFile:

    def __init__(self, config_file_path: str, config_file_content: Dict[str, Any]) -> None:
        # Path to included configuration yml file
        self.config_file_path: str = config_file_path
        # Content of the included configuration yml file
        self.config_file_content: Dict[str, Any] = config_file_content

    def __copy__(self):
        return self

    def __deepcopy__(self, *_args, **_kwargs):
        return self

    def __hash__(self):
        return hash(self.config_file_path)


# @dataclass
class LocalStorageFileReference:

    def __init__(self, lsf: LocalStorageFile, reference: str, as_dict: bool) -> None:
        self.lsf: LocalStorageFile = lsf
        self.reference: str = reference
        self.as_dict: bool = as_dict

    def __copy__(self):
        return self

    def __deepcopy__(self, *_args, **_kwargs):
        return self

    def __hash__(self):
        return hash((self.lsf, self.reference, self.as_dict))


class LocalStorage:

    def __init__(self, local_segment_in_storage: List, local_storage_file: LocalStorageFileReference) -> None:
        """
        :param local_segment_in_storage:  Path from the root configuration to this local storage.
        :param local_storage_file:        Associated storage file
        """
        self.local_segment_in_storage: List = local_segment_in_storage
        self.local_storage_file: LocalStorageFileReference = local_storage_file
        self._configuration_input: Union[List, Dict] = Item.get(self.local_storage_file.lsf.config_file_content,
                                                                local_segment_in_storage)

        self.descendants: List[LocalStorage] = []
        ignored_paths: List[List[str]] = []
        for path, entity in Item.walk_leaves_and_branches(self._configuration_input):
            if any(True for ip in ignored_paths if path[:len(ip)] == ip):
                continue

            if isinstance(entity, LocalStorageFileReference):
                storage_representation: LocalStorageFileReference = entity

                # Path to the include directive in the configuration file.
                new_local_segment: List[str] = [*self.local_segment_in_storage, *path]

                # Parent structure, with  `new_configuration_input[new_local_segment[-1]]` yielding hte LocalStorageFile
                new_configuration_input = Item.get(self._configuration_input, path[:-1])
                assert not len(path) and new_configuration_input is storage_representation or new_configuration_input[path[-1]] is storage_representation

                if isinstance(new_configuration_input, list):
                    # otherwise added as children.
                    if len(path) > 1:
                        self.descendants.append(LocalStorage(new_local_segment[:-1], self.local_storage_file))
                        ignored_paths.append(path[:-1])

                else:
                    self.descendants.append(LocalStorage(new_local_segment, storage_representation))

        self.children: Dict[str, LocalStorage] = {}
        if isinstance(self._configuration_input, list):
            self.children = {
                lsfr.lsf.config_file_path: LocalStorage(self.local_segment_in_storage, lsfr)
                for lsfr in self._configuration_input if isinstance(lsfr, LocalStorageFileReference)}

    def get_data(self) -> Union[Dict, List]:
        data = copy.deepcopy(self._configuration_input)

        # Get data from descendants
        for ls in self.descendants:
            lgd = ls.get_data()
            path = ls.local_segment_in_storage[len(self.local_segment_in_storage):]
            data = uni_set(data, path, lgd)

        # Get data from children
        if isinstance(data, list):

            is_dict = any((d for d in self.children.values() if d.local_storage_file.as_dict))

            r_data = {} if is_dict else []
            for d in data:
                if isinstance(d, LocalStorageFileReference):
                    extracted_data = self.children[d.lsf.config_file_path].get_data()
                    if is_dict:
                        r_data.update(extracted_data)
                    else:
                        r_data.extend(extracted_data)

                else:
                    if is_dict:
                        for path, entity in Item.walk_leaves(d):
                            Item.set(r_data, path, entity)

                        # r_data.update(d)
                    else:
                        r_data.append(d)

            data = r_data

        return data

    def save_data(self, new_data) -> None:

        old_data = self.get_data()

        if len(self.children):
            data_children = [(c, c.get_data()) for c in self.children.values()]
            extends_dict = isinstance(old_data, dict) or isinstance(new_data, dict)

            # Assemble all the override values in one dict
            override_shell = {}
            for d in self._configuration_input:
                if not isinstance(d, LocalStorageFileReference):
                    override_shell.update(d)

            if extends_dict:

                # Editing this does not really work well, if a change occurs, changes are
                # made in the importing configuration file.
                # TODO: Change this in the future, this should be more or less straight forward by iterating
                #       through the children in reverse order, and gradually extending a list of diffs that do not
                #       need to be applied because the keys are already taken from the `extend` child.

                # Remove old content that was removed or changed
                for k, ov in old_data.items():

                    if new_data.get(k) != ov:
                        found: bool = False
                        exists: bool = new_data.get(k) is not None
                        for _child, child_data in data_children:
                            nd_child = copy.deepcopy(new_data)

                            # Make sure not to change the values that are manually overridden in the manual config section
                            override_shell_root = override_shell.get(k)
                            if isinstance(override_shell_root, dict):
                                for path, entity in Item.walk_leaves(override_shell_root):
                                    extended_path = [k, *path]
                                    nd_child = uni_set(nd_child, extended_path, Item.get(child_data, extended_path))

                            if k in child_data and child_data.get(k) != nd_child.get(k):
                                if k not in new_data:
                                    child_data.remove(k)
                                else:
                                    child_data[k] = nd_child.get(k)
                                    found = True

                        # if exists & found, no need to "update" again.
                        if not exists or not found:
                            for cip in self._configuration_input:
                                if isinstance(cip, LocalStorageFileReference):
                                    continue
                                if not exists:
                                    cip.remove(k)
                                else:
                                    cip[k] = new_data.get(k)

                # Commit remove old content
                for child, child_data in data_children:
                    child.save_data(child_data)

                # Add new content to own content
                for k, nv in new_data.items():
                    if k not in old_data:
                        if len(self._configuration_input) and isinstance(self._configuration_input[-1], dict):
                            self._configuration_input[-1][k] = nv
                        else:
                            self._configuration_input.append({k: nv})

            else:

                # Remove old content that was removed
                for k in old_data:
                    if k not in new_data:
                        for _child, child_data in data_children:
                            if k in child_data:
                                child_data.remove(k)

                        if k in self._configuration_input:
                            self._configuration_input.remove(k)

                # Commit remove old content
                for child, child_data in data_children:
                    child.save_data(child_data)

                # Add new content to own content
                for k in new_data:
                    if k not in old_data:
                        self._configuration_input.append(k)

            new_data = self._configuration_input

        for ls in self.descendants:
            # Call to descendant's save_data if something changed
            old_data_descendant = Item.get(old_data, ls.local_segment_in_storage[len(self.local_segment_in_storage):])
            new_data_descendant = Item.get(new_data, ls.local_segment_in_storage[len(self.local_segment_in_storage):])
            if old_data_descendant != new_data_descendant:
                ls.save_data(new_data_descendant)

            path = ls.local_segment_in_storage[len(self.local_segment_in_storage):]
            new_data = uni_set(new_data, path, Item.get(self.local_storage_file.lsf.config_file_content,
                                                        ls.local_segment_in_storage))

        self._configuration_input = new_data
        self.local_storage_file.lsf.config_file_content = uni_set(self.local_storage_file.lsf.config_file_content,
                                                                  self.local_segment_in_storage,
                                                                  self._configuration_input)

    def str(self, prefix='') -> str:
        this_location = f'{"/".join(self.local_segment_in_storage)}' or '/'
        split_path = f'...{os.path.sep.join(self.local_storage_file.lsf.config_file_path.split(os.path.sep)[-2:])}'

        prefix = f'{prefix} '

        children_and_descendants = '\n'.join(chain(
            (c.str(f'{prefix}c') for c in self.children.values()),
            (d.str(f'{prefix}d') for d in self.descendants)))

        if children_and_descendants:
            return f'{prefix}{this_location}{split_path}\n{children_and_descendants}'
        else:
            return f'{prefix}{this_location}{split_path}'

    def flatten(self, data):
        self._configuration_input = self.local_storage_file.lsf.config_file_content = data
        self.descendants: List[LocalStorage] = []


def uni_set(parent, path, value):
    if not len(path):
        return value

    if value is None:
        Item.drop(parent, path)
    else:
        Item.set(parent, path, value)

    return parent
