import logging
import sys
import traceback
import copy
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Union

from dataspree.inference.core.config.configurable import Configurable
from dataspree.inference.core.decorators import Action
from dataspree.inference.core.exceptions import ImplementationError
from dataspree.inference.core.consumer import Consumer
from dataspree.inference.core.item import Item
from dataspree.inference.core.producer import Producer, Property, PropertyType
from dataspree.inference.core.protection import LicenseProtection
from dataspree.inference.core.signal_handler import SignalHandler
from dataspree.inference.core.signals import TerminatedNodeSignal, Signal, StartedNodeSignal
from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.killable_thread import Thread


class Node(SignalHandler, Configurable):
    __application: Optional['Application'] = None

    @staticmethod
    def register(application: 'Application'):
        Node.__application = application

    def __init__(self, name: str, register_signal_producer: bool = False,
                 register_signal_consumer: bool = False) -> None:
        Configurable.__init__(self)
        SignalHandler.__init__(self)
        self.name: str = name

        self.logger = logging.getLogger(f'{type(self).__name__}:{self.name}')

        self._actions: List[str] = [n for n in dir(self.__class__) if isinstance(getattr(self.__class__, n), Action)]
        self._consumers: List[Consumer] = []
        self._producers: List[Producer] = []
        self._running: bool = False
        self._thread: Optional[Thread] = None

        LicenseProtection.raise_node_instantiation_disallowed(self.get_config_key())

        self.direct_signal_producer: Optional[Producer] = None
        self.direct_signal_consumer: Optional[Consumer] = None
        if register_signal_producer:
            self.direct_signal_producer = Producer(f'{name}_signals', [Property('signal', PropertyType.ANY)])
            self.add_producer(self.direct_signal_producer)

        if register_signal_consumer:
            self.direct_signal_consumer = Consumer(f'{name}_signals', signal_handler=self)
            self.add_consumer(self.direct_signal_consumer)

    #
    # Attributes
    #

    @property
    def consumers(self) -> Dict[str, Consumer]:
        return {consumer.name: consumer for consumer in self._consumers}

    @property
    def producers(self) -> Dict[str, Producer]:
        return {producer.name: producer for producer in self._producers}

    @property
    def running(self) -> bool:
        return self._running and self._thread.is_alive()

    @property
    def actions(self) -> List[str]:
        return copy.copy(self._actions)

    def is_active(self) -> bool:
        return self.running

    def __str__(self):
        return f'{self.name}:{type(self).__name__}'

    #
    # Thread management
    #

    def initialize_thread(self) -> None:
        if self._thread is not None and self._thread.is_alive():
            self.logger.warning('Node already started. Stopping and restarting!')
            self.stop()
            self.join(5.0, 5.0)

        if self._thread is None or not self._thread.is_alive():
            def target():
                try:
                    self._running = True
                    self.propagate_signal(StartedNodeSignal(self))
                    self.logger.info('Node started.')
                    self.run()

                except SystemExit:
                    self.logger.error(f'Node {self.name} attempted to stop inference. Intercepted and forwarded.')
                    if self.__application is not None:
                        self.__application.stop()

                finally:
                    self._running = False
                    self.propagate_signal(TerminatedNodeSignal(self))

            self._thread: Thread = Thread(name=self.name, target=target, daemon=True)

        else:
            self.logger.warning('Node thread could not be recreated. Still not terminated.')

    def start(self) -> None:
        self.initialize_thread()
        self._thread.start()

    def stop(self) -> None:
        """Mark the node stopped."""
        self._running = False

    def join(self, timeout_seconds: Optional[float], timeout_ungraceful_stop_seconds: float) -> bool:
        if self._thread is None:
            return True

        try:
            self._thread.join(timeout=timeout_seconds)
        except RuntimeError:
            # If the thread is not yet started
            pass

        if self._thread.is_alive():
            self._thread.stop_ungracefully(timeout_seconds=timeout_ungraceful_stop_seconds)

            if not self._thread.is_alive():
                self.logger.warning(f' > Joined Thread for node {self.name}. Joined thread forcefully.')
                return True

            else:
                self.logger.error(f' > Could not join thread for node {self.name} gracefully within '
                                  f'{timeout_ungraceful_stop_seconds} ({timeout_seconds}) seconds or forcefully '
                                  f'within {timeout_ungraceful_stop_seconds} seconds.')
                return False

        else:
            self.logger.info(f' > Joined Thread for node {self.name}.')
            return True

    #
    # Action API
    #

    def execute_action(self, action_name: str, *args: Any, **kwargs: Any) -> None:
        try:
            func: Optional[Callable] = getattr(self, action_name)
            if func is None:
                raise AttributeError(f'Attribute {action_name} is None.')
            try:
                rv = func(*args, **kwargs)
                self.logger.info(f'Executed action {action_name}')
                return rv
            except Exception as e:
                if __debug__ or is_application_debug():
                    self.logger.warning(f'An exception occurred in action {action_name}: {e}')

                    if is_application_debug():
                        traceback.print_exc()
                else:
                    self.logger.warning(f'An exception occurred in action {action_name}.')

        except AttributeError:
            self.logger.info(f'Attempting to execute action {action_name}, but that action is not registered.')

    #
    # Signal API
    #

    def propagate_signal(self, signal: Signal) -> None:
        for producer in self.producers.values():
            try:
                producer.send_signal(signal)
            except Exception as e:
                raise e from None

    #
    # Producer / consumer API
    #

    def add_consumer(self, consumer: Consumer) -> None:
        if Node.__application is not None:
            Node.__application.register_consumer(consumer, self._consumers)

    def remove_consumer(self, consumer: Consumer) -> None:
        if Node.__application is not None:
            Node.__application.remove_consumer(consumer, self._consumers)

    def add_producer(self, producer: Producer) -> None:
        if Node.__application is not None:
            Node.__application.register_producer(producer, self._producers)

    def remove_producer(self, producer: Producer) -> None:
        if Node.__application is not None:
            Node.__application.remove_producer(producer, self._producers)

    def disconnect_all_producers(self):
        for producer in self._producers:
            producer.disconnect_all_consumers()

    def disconnect_all_consumers(self):
        for consumer in self._consumers:
            consumer.disconnect_all_producers()

    def disconnect(self):
        self.disconnect_all_consumers()
        self.disconnect_all_producers()

    #
    # Execution
    #

    @abstractmethod
    def run(self) -> None:
        """Run the processing loop. Don't call this!"""

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        """Return the properties with which this instances of this Node can be configured."""
        return dict()

    @staticmethod
    @abstractmethod
    def get_config_categories() -> List[str]:
        """Return the categories of the configuration of this kind of nodes."""

    @classmethod
    def get_config_key(cls) -> str:
        """
        Node implementations can implement this function by providing the configuration key
        that identifies their class in the configuration file.
        """
        return super(Node, cls).get_config_key()

    @classmethod
    def build(cls, config: Union[Dict, Item], name: Optional[str], from_user_input: bool = False) \
            -> Optional['Node']:
        """Create an instance of the node from a configuration."""
        return cls.default_build(config=config, name=name, from_user_input=from_user_input)  #

    def terminate_application(self):
        self.logger.warning(f'Node {self} requests to terminate application.')
        sys.exit(1)
