from __future__ import annotations

import datetime
import threading
from typing import List, TypeVar, Generic, Any, TYPE_CHECKING

_SourceT = TypeVar('_SourceT', bound='SignalHandler')
_T = TypeVar('_T')


def non_copyable(cls):
    """Decorator to make class instances only return themselves when copied or deep-copied."""
    cls.__copy__ = lambda self: self
    cls.__deepcopy__ = lambda self, memo: self
    return cls


@non_copyable
class SignalSource(Generic[_SourceT]):
    """A wrapper class for a signal source that should be used for any vertex in the signal routing graph.

    It is used to prevent copying the source instance when the signal is transmitted.
    """

    def __init__(self, source: _SourceT) -> None:
        self._source_instance = source
        self._timestamp = datetime.datetime.now()

    @property
    def source_instance(self) -> _SourceT:
        return self._source_instance

    @property
    def timestamp(self) -> datetime.datetime:
        return self._timestamp

    def __eq__(self, other: Any):
        return isinstance(other, SignalSource) and self._source_instance == other.source_instance

    def __str__(self):
        return str(self._source_instance)


class BaseSignal(Generic[_SourceT]):
    """Base class for all signals that can be transmitted. It does not require an initial source.

    How a signal is processed and propagated is defined in the SignalHandler implementation.
    If a new initialized signal instance should be emitted, the propagate_signal method of the SignalHandler can be used.
    """
    __bs_id = 0
    __lock = threading.Lock()

    def __init__(self):
        self._routing: List[SignalSource[_SourceT]] = []

        # Note when module should be converted to C (like with Cython):
        # Large numbers in the python int object are not be handled correspondingly after cythonizing.
        # In python, as long as memory is available, integers of arbitrary size can be incremented.
        #
        # In C, the max possible integer depends on the C compiler and machine architecture.
        # E.g. counter not a problem for 64 bit systems, for 32-bit targets and cythonization, a different
        # implementation should be used. (I.e. storing a not-copyable self reference self.r and signal_id := id(self.r))
        # This way, the original signal object is prevented from being copied and the signal_id is unique during the
        # lifetime. Drawback: the id is not successive anymore.
        with BaseSignal.__lock:
            # thread-safe signal id generation
            BaseSignal.__bs_id += 1
            self.__signal_id = BaseSignal.__bs_id

    @property
    def signal_id(self) -> int:
        return self.__signal_id

    def loop_finished(self, source: _SourceT) -> bool:
        for node in self._routing:
            if node.source_instance == source:
                return True

        return False

    def transmit(self, source: _SourceT):
        """Wraps the propagation of the signal by adding signal transmission information
        to the signal and preventing cyclic routing.

        This function is used by the SignalHandler internally and is not supposed to be called directly.
        For emitting a new signal, the propagate_signal method of the SignalHandler should be used instead.

        :param source: the source (SignalHandler) instance that is supposed to transmit the signal.
        """
        for node in self._routing:
            if node.source_instance == source:
                return
        self._routing.append(SignalSource(source))
        source.propagate_signal(self)

    def __str__(self):
        route_str = '->'.join((str(r) for r in self._routing))
        return f'{type(self).__name__} ({self.signal_id}) via {route_str}'


class Signal(Generic[_SourceT], BaseSignal[_SourceT]):
    """A signal which is initially emitted by a source."""

    def __init__(self, source: _SourceT):
        BaseSignal.__init__(self)
        self._routing.append(SignalSource(source))

    @property
    def _source(self) -> _SourceT:
        return self._routing[0].source_instance


class NodeSignal(Signal['Node']):
    """A signal which is propagated via inference nodes (dataspree.inference.core.node.Node).

    How a signal is processed and propagated is defined in the SignalHandler implementation.
    If a new initialized signal instance should be emitted, the propagate_signal method of the SignalHandler can be used.
    """

    @property
    def source_type(self):
        return type(self._source)

    @property
    def source_name(self):
        return self._source.name


class StartedNodeSignal(NodeSignal):
    pass


class TerminatedNodeSignal(NodeSignal):
    pass


class EndOfFileSignal(NodeSignal):
    pass
