import datetime
import logging
import time
from typing import Optional, Dict, Any

import numpy as np

from dataspree.inference.core.config import ConfigFieldTypes
from dataspree.inference.core.consumer import Consumer
from dataspree.inference.core.decorators import action
from dataspree.inference.core.item import Item
from dataspree.inference.core.plugin import Plugin
from dataspree.inference.core.producer import Producer

logger = logging.getLogger(__name__)


class QualityInspectionPlugin(Plugin):
    def __init__(self, name: str, ok_class_id: int) -> None:
        super(QualityInspectionPlugin, self).__init__(name)

        self.consumer: Consumer = Consumer(name)
        self.add_consumer(self.consumer)

        self.producer: Producer = Producer(name)
        self.add_producer(self.producer)

        self.ok_class_id: int = ok_class_id

        self.ok_count: int = 0
        self.nok_count: int = 0

        self.pause: bool = False
        self.latest_item: Optional[Item] = None

        self.send_update_on_next_tick: bool = False

    @staticmethod
    def get_class_label_and_confidence(item):
        class_label = item['inference']['classification']['label']
        confidence = item['inference']['classification']['confidence']

        return class_label, confidence

    def run(self) -> None:
        while self.running:
            item: Optional[Item] = self.consumer.receive()
            if item is None or self.pause:
                if self.send_update_on_next_tick:
                    self._send_update(include_item=False)
                time.sleep(0.001)
                continue

            class_label, confidence = self.get_class_label_and_confidence(item)
            is_ok = class_label == self.ok_class_id

            if is_ok:
                self.ok_count += 1
            else:
                self.nok_count += 1

            # Info for current item
            info = dict(
                result='OK' if is_ok else 'NOK',
                result_variant='success' if is_ok else 'danger',
                identifier=str(item['id']),
                confidence=confidence,
                datetime=datetime.datetime.fromtimestamp(item['timestamp'] / 1000).isoformat(),
            )

            self.latest_item = Item(
                id=item['id'],
                timestamp=item['timestamp'],
                image=np.array(item['image']),
                info=info
            )

            new_item = Item(
                id=item['id'],
                timestamp=item['timestamp'],
                image=np.array(item['image']),
                info=info
            )
            self.latest_item = new_item
            self._send_update(include_item=True)

    @action
    def toggle_pause(self) -> None:
        self.pause = not self.pause
        self.send_update_on_next_tick = True

    @action
    def start_action(self) -> None:
        self.pause = False
        self.send_update_on_next_tick = True

    @action
    def stop_action(self) -> None:
        self.pause = True
        self.send_update_on_next_tick = True

    @action
    def reset_statistics(self) -> None:
        self.ok_count = 0
        self.nok_count = 0
        self.send_update_on_next_tick = True

    def _send_update(self, include_item: bool = False) -> None:
        self.send_update_on_next_tick = False

        if include_item and self.latest_item is not None:
            update_item = self.latest_item
        else:
            update_item = Item()

        error_rate = '-'
        if self.ok_count + self.nok_count > 0:
            error_rate = f'{100 * self.nok_count / (self.ok_count + self.nok_count):.4g}%'

        # Calculate KPIs
        update_item['kpis'] = dict(
            total=str(self.ok_count + self.nok_count),
            error_rate=error_rate,
            ok=str(self.ok_count),
            nok=str(self.nok_count),
        )

        # Determine action visibility
        update_item['action_visibility'] = dict(
            start_visible=self.pause,
            stop_visible=not self.pause,
        )

        self.producer.send(update_item)

    @classmethod
    def get_config_entries(cls) -> Dict[str, Any]:
        return {
            'display_name': 'Quality Inspection',
            'description': 'Generic quality inspection plugin that can be used in combination with the quality '
                           'inspection UI. You only need to specify the OK class ID to use this plugin. '
                           'Statistics will automatically be collected',
            'fields': [
                {
                    'display_name': 'OK class ID',
                    'name': 'ok_class_id',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                }
            ]
        }
