from dataclasses import dataclass, field

from typing import List, Union, Optional, Set

from dataspree.inference.sensor.data.pixel_format import PixelFormat
from dataspree.inference.sensor.data.sensor_model import SensorModel


@dataclass
class CameraModel(SensorModel):
    """
    Arguments:

        available_pixel_formats (List[PixelFormat]):
            Supported pixel formats.

        available_configuration_names (List[str])
            The names of writable configurations.

        available_configuration_names_read_only (List[str])
            The names of factory settings that cannot be adjusted / saved to.

        acquisition_burst_modes (List[str]):
            Modes for acquisition bursts.

        white_balance_manual_channels (List[Union[str, int]]):
            Channels for manual white balance adjustment.

        rois (List[str]):
            Regions of interest.

        auto_function_profiles (List[str]):
            Profiles for automatic functions.

        gpio_lines (List[str]):
            General-purpose input/output lines.

        software_trigger_sources_generic (List[str]):
            Generic software trigger sources.

        software_trigger_frame_start (Optional[str]):
            Software trigger for frame start.

        available_channels (List[str]):
            Available channels.

        binning_modes (List[str]):
            Modes for binning.

        contrast_modes (List[str]):
            Modes for contrast adjustment.

        supported_denoising_modes (Optional[List[str]]):
            Supported denoising modes.

        supported_demosaicing_modes (List[str]):
            Supported demosaicing modes.

        frame_start_trigger (Optional[str]):
            Trigger for frame start.

        frame_stop_trigger (Optional[str]):
            Trigger for frame stop.

        frame_active_trigger (Optional[str]):
            Trigger for frame active.

        frame_burst_start_trigger (Optional[str]):
            Trigger for frame burst start.

        frame_burst_stop_trigger (Optional[str]):
            Trigger for frame burst stop.

        frame_burst_active_trigger (Optional[str]):
            Trigger for frame burst active.

        exposure_start_trigger (Optional[str]):
            Trigger for exposure start.

        exposure_stop_trigger (Optional[str]):
            Trigger for exposure stop.

        exposure_active_trigger (Optional[str]):
            Trigger for exposure active.

        line_start_trigger (Optional[str]):
            Trigger for line start.




    """
    available_pixel_formats: List[PixelFormat] = field(default_factory=list)
    available_configuration_names_writable: List[str] = field(default_factory=list)
    available_configuration_names_read_only: List[str] = field(default_factory=list)
    available_configuration_names: List[str] = field(default_factory=list)

    acquisition_burst_modes: List[str] = field(default_factory=list)

    white_balance_manual_channels: List[Union[str, int]] = field(default_factory=list)
    rois: List[str] = field(default_factory=list)
    auto_function_profiles: List[str] = field(default_factory=list)

    gpio_lines: List[str] = field(default_factory=list)

    software_trigger_sources_generic: List[str] = field(default_factory=list)
    software_trigger_frame_start: Optional[str] = 'Software'

    available_channels: List[str] = field(default_factory=list)
    binning_modes: List[str] = field(default_factory=list)
    contrast_modes: List[str] = field(default_factory=list)

    supported_denoising_modes: Optional[List[str]] = field(default_factory=list)
    supported_demosaicing_modes: List[str] = field(default_factory=list)

    # TODO: Update through image_acquirer.remote_device.node_map.TriggerSelector.symbolics
    # TODO: DeviceScanType contains information on whether this is a lince scanner or area scan device.
    frame_start_trigger: Optional[str] = None
    frame_stop_trigger: Optional[str] = None
    frame_active_trigger: Optional[str] = None

    frame_burst_start_trigger: Optional[str] = None
    frame_burst_stop_trigger: Optional[str] = None
    frame_burst_active_trigger: Optional[str] = None

    exposure_start_trigger: Optional[str] = None
    exposure_stop_trigger: Optional[str] = None
    exposure_active_trigger: Optional[str] = None

    line_start_trigger: Optional[str] = None

    def __post_init__(self) -> None:
        if not len(self.available_configuration_names):
            self.available_configuration_names = (self.available_configuration_names_writable +
                                                  self.available_configuration_names_read_only)

    @property
    def trigger_sources_generic(self) -> List[str]:
        return self.gpio_lines + self.software_trigger_sources_generic

    @property
    def all_trigger_names(self) -> Set[str]:
        return self.acquisition_trigger_names | self.exposure_trigger_names

    @property
    def acquisition_trigger_names(self) -> Set[str]:
        atn = {self.frame_start_trigger, self.frame_stop_trigger, self.frame_active_trigger,
               self.frame_burst_start_trigger, self.frame_burst_stop_trigger, self.frame_burst_active_trigger, }
        atn.discard(None)
        return atn

    @property
    def exposure_trigger_names(self) -> Set[str]:
        etn = {self.exposure_start_trigger, self.exposure_stop_trigger, self.exposure_active_trigger}
        etn.discard(None)
        return etn
