from abc import ABCMeta, abstractmethod

from enum import Enum
from typing import List, Union, Tuple, Dict, Optional, get_args
from logging import getLogger

from dataspree.inference.sensor.data.pixel_format.utils import all_implementations

logger = getLogger(__name__)


class DefaultDataType(Enum):
    UNSIGNED = ''  # resp. u
    SIGNED = 's'
    FLOAT = 'f'


_custom_data_types: List[str] = []

DataType = Union[DefaultDataType, str]


def is_data_type(potential_data_type):
    for arg in get_args(DataType):
        if isinstance(potential_data_type, arg):
            return True

    return False


_reverse_data_type_mapping: Dict[str, DataType] = {f.value: f for f in DefaultDataType}
_reverse_data_type_mapping['u'] = DefaultDataType.UNSIGNED


def datatype_name(data_type: DataType, multiple_designators: bool = False) -> str:
    if isinstance(data_type, DefaultDataType):
        if multiple_designators and data_type is DefaultDataType.UNSIGNED:
            return 'u'
        return data_type.value
    return data_type


def register_custom_data_type(name: str) -> None:
    if name in _custom_data_types:
        raise ValueError
    if name in _reverse_data_type_mapping:
        raise ValueError
    if len(name) != 1:
        raise ValueError
    _custom_data_types.append(name)


def data_types() -> List[DataType]:
    return list(_custom_data_types) + list(DefaultDataType)


class DataTypes(metaclass=ABCMeta):

    @abstractmethod
    def designation(self) -> str:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    @abstractmethod
    def __getitem__(self, item):
        ...

    @staticmethod
    def create(full_designation: str, number_of_channels: int) -> List[Tuple['DataTypes', str]]:
        compos: List[Tuple[DataTypes, str]] = []
        for implementation in all_implementations(DataTypes):
            try:
                compos.extend(implementation.from_designation(full_designation, number_of_channels))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str, number_of_channels: int) -> List[Tuple['DataTypes', str]]:
        ...


class CompoundDataTypes(DataTypes):
    def __init__(self, data: Optional[Union[List[DataType], DataType]]) -> None:
        if data is None:
            data = DefaultDataType.UNSIGNED

        if is_data_type(data):
            data = [data]

        self.data: List[DataType] = data

    def designation(self) -> str:
        if not len(self.data):
            return ''

        if len(set(self.data)) == 1:
            return datatype_name(self.data[0])

        return ''.join((datatype_name(s) for s in self.data))

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def __eq__(self, other):
        return isinstance(other, CompoundDataTypes) and self.data == other.data

    @staticmethod
    def from_designation(full_designation: str, number_of_channels: int) -> List[Tuple[DataTypes, str]]:
        # Compound data type names don't contain numbers
        possible_data_types: List[Tuple[DataTypes, str]] = [(CompoundDataTypes(DefaultDataType.UNSIGNED),
                                                             full_designation)]

        # Is the first letter a data type?

        decoded_data_types: List[DataType] = []
        for n in range(number_of_channels):
            if n >= len(full_designation):
                break

            data_type_designation: str = full_designation[n]
            if data_type_designation in _reverse_data_type_mapping:
                decoded_data_types.append(_reverse_data_type_mapping[data_type_designation])
            else:
                break

        if len(decoded_data_types) == number_of_channels:
            possible_data_types.append((CompoundDataTypes(decoded_data_types), full_designation[number_of_channels:]))

        if len(decoded_data_types) > 1:
            possible_data_types.append((CompoundDataTypes(decoded_data_types[0]), full_designation[1:]))

        return possible_data_types
