from abc import ABCMeta, abstractmethod
from logging import getLogger

from typing import List, Tuple, Dict

from dataspree.inference.sensor.data.pixel_format.number_of_bits import Size
from dataspree.inference.sensor.data.pixel_format.components import (channel_name, Components,
                                                                     LMNLocation, GenicamComponents)
from dataspree.inference.sensor.data.pixel_format.utils import all_implementations

logger = getLogger(__name__)


class InterfaceSpecific(metaclass=ABCMeta):

    @abstractmethod
    def designation(self) -> str:
        ...

    @staticmethod
    def create(full_designation: str, components: Components,
               sizes: Size) -> List[Tuple['InterfaceSpecific', str]]:

        if not full_designation.startswith('_'):
            return []

        full_designation = full_designation[1:]

        compos: List[Tuple[InterfaceSpecific, str]] = []
        for implementation in all_implementations(InterfaceSpecific):
            try:
                compos.extend(implementation.from_designation(full_designation, components, sizes))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str, components: Components,
                         sizes: Size) -> List[Tuple['InterfaceSpecific', str]]:
        ...


class Planar(InterfaceSpecific):

    def designation(self) -> str:
        return 'Planar'

    @staticmethod
    def from_designation(full_designation: str, components: Components,
                         sizes: Size) -> List[Tuple['InterfaceSpecific', str]]:
        designation = 'Planar'
        if full_designation.startswith(designation):
            return [(Planar(), full_designation[len(designation):])]
        return []

    def __eq__(self, other) -> bool:
        return isinstance(other, Planar)


class SemiPlanar(InterfaceSpecific):

    def designation(self) -> str:
        return 'Semiplanar'

    @staticmethod
    def from_designation(full_designation: str, components: Components,
                         sizes: Size) -> List[Tuple['InterfaceSpecific', str]]:
        designation = 'Semiplanar'
        if full_designation.startswith(designation):
            return [(Planar(), full_designation[len(designation):])]
        return []

    def __eq__(self, other) -> bool:
        return isinstance(other, SemiPlanar)


class ComponentsSequencing(InterfaceSpecific):

    def __init__(self, components: Tuple[str, ...]) -> None:
        self.components: Tuple[str, ...] = components

    def designation(self) -> str:
        return ''.join(self.components)

    @staticmethod
    def from_designation(full_designation: str, components: Components,
                         sizes: Size) -> List[Tuple['InterfaceSpecific', str]]:
        component_names: List[str] = [channel_name(components[i]) for i in range(len(components))]
        if not len(component_names):
            return []

        component_counts: Dict[int, int] = {}
        extracted_components: List[str] = []

        i_f: int = 0
        while i_f <= len(full_designation):
            substr = full_designation[i_f:]

            matching_component_index: int
            matching_component_name: str
            for matching_component_index, matching_component_name in enumerate(component_names):
                if substr.startswith(matching_component_name):
                    break
            else:
                break

            extracted_components.append(matching_component_name)
            component_counts.setdefault(matching_component_index, 0)
            component_counts[matching_component_index] += 1

            i_f += len(matching_component_name)

        if len(component_counts) != len(sizes):
            return []

        if isinstance(components, GenicamComponents) and isinstance(components.location, LMNLocation):
            int_occurrences = components.location.fields
            if len(int_occurrences) != len(sizes):
                return []

            max_occurrence: int = max(int_occurrences)
            max_component_counts: int = max(component_counts.values())

            for i_c, int_occurrence in enumerate(int_occurrences):
                o_adjusted = int_occurrence * max_component_counts
                count_adjusted = component_counts[i_c] * max_occurrence

                if o_adjusted != count_adjusted:
                    return []

        return [(ComponentsSequencing(tuple(extracted_components)), full_designation[i_f:])]

    def __eq__(self, other) -> bool:
        return isinstance(other, ComponentsSequencing) and self.components == other.components
