import numpy as np

from abc import abstractmethod, ABCMeta
from typing import Union, List, Optional, Tuple

from logging import getLogger

from dataspree.inference.sensor.data.exceptions import FeatureImplementationError, FeatureNotImplemented
from dataspree.inference.sensor.data.pixel_format.components import Components, SizedColorSpaceStandard

from dataspree.inference.sensor.data.pixel_format.data_types import DataTypes, CompoundDataTypes, DataType
from dataspree.inference.sensor.data.pixel_format.interface_specifics import InterfaceSpecific
from dataspree.inference.sensor.data.pixel_format.number_of_bits import Size
from dataspree.inference.sensor.data.pixel_format.utils import join_designators, all_implementations
from dataspree.inference.sensor.data.pixel_format.packing_types import Packing

logger = getLogger(__name__)


class PixelFormat(metaclass=ABCMeta):
    """
    Abstraction for arbitrary pixel formats.
    """

    @staticmethod
    def create(full_designation: str) -> List['PixelFormat']:
        compos: List[PixelFormat] = []
        for implementation in all_implementations(PixelFormat):
            try:
                compos.extend(implementation.from_designation(full_designation))
            except Exception as e:
                logger.warning(f'An exception occurred in Pixel Format ' f'implementation {implementation.__name__}:'
                               f' {e}')
        return compos

    @staticmethod
    def from_name(full_designation: str) -> 'PixelFormat':
        matching_formats: List[PixelFormat] = PixelFormat.create(full_designation)
        if not len(matching_formats):
            raise FeatureNotImplemented(f'Cannot find a registered pixel format "{full_designation}".')

        if len(matching_formats) != 1:
            class_names: List[str] = [type(f).__name__ for f in matching_formats]
            raise FeatureImplementationError(
                f'Cannot find a registered pixel format {full_designation}. Implementations found:'
                f'{", ".join(class_names)}')

        return matching_formats[0]

    @abstractmethod
    def designation(self) -> str:
        ...

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str) -> List['PixelFormat']:
        ...

    @abstractmethod
    def get_structured_array(self, channels: List[np.ndarray]) -> Optional[np.ndarray]:
        ...


class GenicamPixelFormat(PixelFormat):
    """
    Detailed information about pixel formats can be found in the GenICam Pixel Format Naming Convention 2.1.
    https://www.emva.org/wp-content/uploads/GenICam_PFNC_2_1.pdf

    The naming convention subdivides as it follows:
    1. Components and Location (e.g., "RGB" / "Mono")
    2. number of bits per component (e.g., "8")
    3. [Optional] data type ("u" or "" for unsigned, "s" for signed, "f" for floating point)
    4. [Optional] packing: @see Packing
    5. [Optional] interface-specific

    Consecutive fields must be separated by an underscore, iff (:= if, and only if) a pixel name
    requires 2 numbers in its designation as part of consecutive fields.
    Example: YCbCr709_422_8 for 8-bit per component Y’CbCr 4:2:2 using ITU-R BT.709.

    """

    def __init__(self, component: Components, size: Union[Size, int, List[int]],
                 data_type: Optional[Union[DataTypes, List[DataType], DataType]] = None,
                 packing: Optional[Packing] = None,
                 interface_specific: Optional[InterfaceSpecific] = None,
                 sized_color_space_standard: Optional[SizedColorSpaceStandard] = None) -> None:
        if isinstance(size, int):
            size = Size([size, ] * len(component))
        elif isinstance(size, list):
            size = Size(size)

        if not isinstance(data_type, DataTypes):
            data_type = CompoundDataTypes(data_type)

        if not isinstance(size, Size):
            raise ValueError

        self.component: Components = component
        self.size: Size = size
        self.sized_color_space_standard: Optional[SizedColorSpaceStandard] = sized_color_space_standard
        self.data_type: DataTypes = data_type
        self.packing: Packing = packing or Packing()
        self.interface_specific: Optional[InterfaceSpecific] = interface_specific
        self.__designation: Optional[str] = None

    def designation(self) -> str:
        if self.__designation is not None:
            return self.__designation

        comp: str = self.component.full_designation()
        size: str = self.size.designation()
        sized_css: str = self.sized_color_space_standard.name if self.sized_color_space_standard else ''
        data: str = self.data_type.designation()
        pack: str = self.packing.designation()
        descriptor: str = join_designators(join_designators(join_designators(join_designators(
            comp, size), sized_css), data), pack)

        if self.interface_specific is not None:
            self.__designation = f'{descriptor}_{self.interface_specific.designation()}'
        else:
            self.__designation = descriptor

        return self.__designation

    def __setattr__(self, key, value):
        initialized: bool
        try:
            initialized = self.__designation or True
        except AttributeError:
            initialized = False

        if initialized and not key.startswith('_'):
            self.__designation = None

        PixelFormat.__setattr__(self, key, value)

    @staticmethod
    def from_designation(full_designation: str) -> List['GenicamPixelFormat']:
        possible_pixel_formats: List[GenicamPixelFormat] = []

        possible_components: List[Tuple[Components, str]] = Components.create(full_designation)
        for component, component_substring in possible_components:

            size_info = Size.from_designation(component_substring, len(component))
            if size_info is None:
                continue
            size, size_substring = size_info

            optional_standards = SizedColorSpaceStandard.create(size_substring, component, size)
            for scs, scs_substring in optional_standards:

                possible_data_types = DataTypes.create(scs_substring, len(component))
                for data_type, data_type_substring in possible_data_types:

                    possible_pack_types = Packing.from_designation(data_type_substring)
                    for pack_type, pack_type_substring in possible_pack_types:

                        if not len(pack_type_substring):
                            possible_pixel_formats.append(GenicamPixelFormat(component, size, data_type,
                                                                             pack_type, None, scs))

                        else:
                            possible_interfaces = InterfaceSpecific.create(data_type_substring, component, size)

                            for interface, remaining_substring in possible_interfaces:
                                if not len(remaining_substring):
                                    possible_pixel_formats.append(GenicamPixelFormat(component, size, data_type,
                                                                                     pack_type, interface, scs))

        return possible_pixel_formats

    def __eq__(self, other) -> bool:
        return (isinstance(other, GenicamPixelFormat) and self.component == other.component and
                self.size == other.size and self.data_type == other.data_type and
                self.packing == other.packing and self.interface_specific == other.interface_specific and
                self.sized_color_space_standard == other.sized_color_space_standard)

    def get_structured_array(self, channels: List[np.ndarray]) -> Optional[np.ndarray]:
        return self.component.get_structured_array(channels)
