from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional, TypeVar, Generic
from abc import ABCMeta, abstractmethod

from dataspree.inference.core.exceptions import ImplementationError
from dataspree.inference.sensor.data.exceptions import InvalidState


class SensorState(Enum):
    """Contains all the states in a sensor livecycle."""
    UNAVAILABLE = 0
    DISCOVERED = 1
    CONNECTED = 2
    STARTED = 3


class SensorStateTransition(Enum):
    """Contains all the state transitions."""
    STOPPED = -2
    DISCONNECTED = -1
    LOST = 0
    FOUND = 1
    CONNECTED = 2
    STARTED = 3


# TODO: Document the intended behavior and unit test it!
class OptionalConfiguration:
    def __init__(self):
        self.__initialized: bool = True

    def entails(self, other) -> bool:
        """
        Returns: If this configuration encompasses the
        """
        if not is_available(other):
            return True

        if not is_available(self):
            return False

        if type(self) is not type(other):
            return False

        for a in vars(self):
            try:
                mine = getattr(self, a)
                others = getattr(other, a)
            except AttributeError:
                return False
            if not is_available(others):
                continue

            try:
                if not mine.entails(others):
                    return False
            except AttributeError:
                if others != mine:
                    return False
        return True

    def __setattr__(self, key, value):
        try:
            initialized = self.__initialized
        except AttributeError:
            initialized = False

        if not initialized:
            super().__setattr__(key, value)
            return

        else:
            if is_available(value):
                return super().__setattr__(key, value)

            previous_value = getattr(self, key, value)
            if previous_value is None:
                return super().__setattr__(key, value)

            # previous value is not None, but might still be unavailable.
            if isinstance(value, NotSupported, ):
                return super().__setattr__(key, value)

            elif isinstance(value, NotRetrieved):
                pass

            elif isinstance(value, NotSpecified):
                pass


# TODO: Document these availability classes
class NotAvailable(OptionalConfiguration):
    pass


class NotAvailableRead(NotAvailable):
    pass


class NotAvailableWrite(NotAvailable):
    pass


class NotSupported(NotAvailableRead):
    pass


class NotRetrieved(NotAvailableRead):
    pass


class TemporarilyNotAvailable(NotRetrieved):
    pass


class NotSpecified(NotAvailableWrite):
    pass


def is_available(value: Any):
    return not isinstance(value, NotAvailable) and value is not None


@dataclass
class Sensor(OptionalConfiguration):
    """Sensor representation.

    Args:
        identifier (Optional[str]):
            Identifier that must be unique. Is used to identify a particular device. Can also be used to determine
            the sensor model if a model for a particular device is registered.
            If this is None, the sensor instance is just used for supplying configurations to one or more actual
            sensors.

        internal_device_id (Optional[int])
            identifier for the driver (if necessary). Might change with new cameras connecting, so we cannot use this
            value reliably outside the driver.

        model (Optional[str]):
            Sensor model. Is determined the best fitting sensor mode.

        manufacturer (Optional[str]):
            Sensor manufacturer. Example: Basler. Is used to determine the best fitting model.

        name (Optional[str]):
            User defined name.

        firmware_version (Optional[str]):
            Firmware version gathered from the device.

        state (SensorState):
            Current state of the sensor.

        device_properties (Dict[str, Any]):
            Additional properties that are relevant to the sensor.
    """
    identifier: Optional[str]
    internal_device_id: Optional[int] = None
    model: Optional[str] = None
    manufacturer: Optional[str] = None
    name: Optional[str] = None
    firmware_version: Optional[str] = None

    state: SensorState = SensorState.UNAVAILABLE
    device_properties: Dict[str, Any] = field(default_factory=dict)
    __callback: Optional['SensorCallback'] = None
    __initialized: bool = field(init=False)

    def __post_init__(self):
        self.__initialized = True

    def register_callback(self, cb: 'SensorCallback') -> None:
        if self.__callback is not None:
            raise InvalidState

        self.__callback = cb

    def __report_sensor_lost(self):
        try:
            self.__callback.on_sensor_lost(self)
        except AttributeError:
            raise ImplementationError

    def __report_sensor_found(self):
        try:
            self.__callback.on_sensor_found(self)
        except AttributeError:
            raise ImplementationError
        super().__setattr__('sensor', SensorState.DISCOVERED)

    def __report_sensor_connect(self):
        try:
            self.__callback.on_sensor_connected(self)
        except AttributeError:
            raise ImplementationError
        super().__setattr__('sensor', SensorState.CONNECTED)

    def __report_sensor_disconnect(self):
        try:
            self.__callback.on_sensor_disconnected(self)
        except AttributeError:
            raise ImplementationError
        super().__setattr__('sensor', SensorState.CONNECTED)

    def __report_sensor_start(self):
        try:
            self.__callback.on_sensor_started(self)
        except AttributeError:
            raise ImplementationError
        super().__setattr__('sensor', SensorState.STARTED)

    def __report_sensor_stop(self):
        try:
            self.__callback.on_sensor_stopped(self)
        except AttributeError:
            import traceback
            traceback.print_exc()
            raise ImplementationError
        super().__setattr__('sensor', SensorState.STARTED)

    def __report_sensor_updated(self):
        try:
            self.__callback.on_sensor_updated(self)
        except AttributeError:
            raise ImplementationError

    def __setattr__(self, key, value) -> None:
        """Prohibit writing the identifier."""
        if key == 'identifier':
            try:
                if self.__initialized and self.identifier is not None:
                    raise ValueError('Cannot change identifier.')

            except AttributeError:
                pass

        if key == 'state':
            try:
                initialized = self.__initialized
            except AttributeError:
                initialized = False

            if initialized:

                for event in range(self.state.value + 1, value.value + 1):
                    try:
                        {SensorState.DISCOVERED.value: self.__report_sensor_found,
                         SensorState.CONNECTED.value: self.__report_sensor_connect,
                         SensorState.STARTED.value: self.__report_sensor_start}[event]()
                    except KeyError:
                        raise ImplementationError

                for event in list(range(value.value, self.state.value))[::-1]:
                    try:
                        {SensorState.UNAVAILABLE.value: self.__report_sensor_lost,
                         SensorState.DISCOVERED.value: self.__report_sensor_disconnect,
                         SensorState.CONNECTED.value: self.__report_sensor_stop}[event]()
                    except KeyError:
                        raise ImplementationError

        OptionalConfiguration.__setattr__(self, key, value)

    def __str__(self):
        return self.identifier


def copy_ignored(cls):
    """Decorator to make class instances only return themselves when copied or deep-copied."""
    cls.__copy__ = lambda self: None
    cls.__deepcopy__ = lambda self, memo: None
    return cls


T = TypeVar('T', bound=Sensor)


@copy_ignored
class SensorCallback(Generic[T], metaclass=ABCMeta):
    @abstractmethod
    def on_sensor_tracked(self, sensor_identifier: str) -> None:
        ...

    @abstractmethod
    def on_sensor_found(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_lost(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_connected(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_disconnected(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_started(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_stopped(self, sensor: T) -> None:
        ...

    @abstractmethod
    def on_sensor_updated(self, sensor: T) -> None:
        ...


@dataclass
class SensorData:
    """Basic sensor data representation

    Arguments:
         timestamp_inference (int): the timestamp of creation / acquisition.
    """
    timestamp_inference: int


class SensorLost(Exception):
    def __init__(self, sensor: Sensor) -> None:
        self.sensor: Sensor = sensor
