from __future__ import annotations

import copy
import threading
import weakref
from logging import getLogger

from typing import Optional, Set, Dict

from harvesters.core import Harvester

from dataspree.inference.sensor.data.exceptions import DriverState

logger = getLogger(__name__)


class HarvesterProxy:
    """A proxy to manage the Harvester instance lifecycle."""

    def __init__(self, harvester: weakref.ref[Harvester]) -> None:
        """
        Initialize the HarvesterProxy with a Harvester instance.

        Args:
            harvester (Harvester): The harvester instance to be managed.
        """
        self.harvester: weakref.ref[Harvester] = harvester

    def destroy(self) -> None:
        """Destroy the Harvester instance by setting it to None."""
        self.harvester = None


class HarvesterLock:
    """A lock manager for Harvester instance to ensure thread-safe operations."""

    def __init__(self, harvester: weakref.ref[Harvester]) -> None:
        """
        Initialize the HarvesterLock with a Harvester instance.

        Args:
            harvester (weakref.ref[Harvester]): The harvester instance to be managed.
        """
        self.__owner_lock: threading.Lock = threading.Lock()
        self.__owners: Dict[int, Set[str]] = {}
        self.__next_driver_slot: int = 0

        self.__harvester_lock: threading.Lock = threading.Lock()
        self.__harvester: weakref.ref[Harvester] = harvester
        self.__harvester_proxy: Optional[HarvesterProxy] = None

    @property
    def owners(self) -> Dict[int, Set[str]]:
        """
        Get a copy of the current owners.

        Returns:
            Dict[int, Set[str]]: A deep copy of the owners.
        """
        with self.__owner_lock:
            return copy.deepcopy(self.__owners)

    def register(self, cti_files: Set[str], update: bool = True) -> int:
        """
        Register a set of CTI files and update the harvester if required.

        Args:
            cti_files (Set[str]): A set of CTI file paths.
            update (bool): Flag to update the harvester if new files are added.

        Returns:
            int: The driver slot assigned to the registered CTI files.
        """
        driver_slot: int
        with self.__owner_lock:
            self.__next_driver_slot += 1
            driver_slot = self.__next_driver_slot
            assert driver_slot not in self.__owners
            self.__owners[driver_slot] = cti_files

        file_added: bool = False
        with self as proxy:
            for cti_file in cti_files:
                if cti_file not in self.__harvester.files:
                    proxy.harvester.add_file(cti_file, check_existence=True, check_validity=True)
                    file_added = True

            if file_added and update:
                proxy.harvester.update()

        return driver_slot

    def deregister(self, slot: int, update: bool = True) -> None:
        """
        Deregister a driver slot and update the harvester if required.

        Args:
            slot (int): The driver slot to be deregistered.
            update (bool): Flag to update the harvester if files are removed.

        Raises:
            ValueError: If the driver slot is not found.
        """
        with self.__owner_lock:
            try:
                cti_files_to_remove: Set[str] = self.__owners.pop(slot)
            except KeyError:
                raise ValueError('Driver not found.')

            for cti_files in self.__owners.values():
                for cti_file in cti_files:
                    cti_files_to_remove.discard(cti_file)

        for cti_file_to_remove in cti_files_to_remove:
            self.__harvester.remove_file(cti_file_to_remove)

        if len(cti_files_to_remove) and update:
            try:
                self.__harvester.update()
            except Exception:
                raise DriverState

    def update(self) -> None:
        """Update the harvester."""
        with self as proxy:
            proxy.harvester.update()

    def __enter__(self) -> HarvesterProxy:
        """
        Enter the context manager, acquiring the harvester lock.

        Returns:
            HarvesterProxy: A proxy to the Harvester instance.

        Raises:
            ValueError: If the lock is already acquired.
        """
        if self.__harvester_proxy is not None:
            raise ValueError('Lock already acquired.')

        self.__harvester_lock.acquire()
        self.__harvester_proxy = HarvesterProxy(self.__harvester)
        return weakref.proxy(self.__harvester_proxy)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Exit the context manager, releasing the harvester lock.
        @dev: Don't access public or protected members here.
        """
        self.__harvester_proxy.destroy()
        self.__harvester_proxy = None
        self.__harvester_lock.release()

    def __getattribute__(self, item):
        # Raise reference error if we try to do stuff even though harvester is out of scope.
        if not item.startswith('_HarvesterLock'):
            try:
                _hash = self.__harvester.__hash__
            except ReferenceError:
                raise ReferenceError('The harvester object is released.') from None

        return object.__getattribute__(self, item)


class Registry:
    """
    Responsible for managing the setup and teardown of the harvester instance.
    """

    def __new__(cls, *args, **kwargs):
        """
        Create a new instance of the registry if it is the first instance.

        Raises:
            Exception: If an instance already exists.

        Returns:
            Registry: The unique instance of the registry.
        """
        instance = getattr(cls, '__instance__', None)
        if instance is not None:
            raise Exception('Use `get_multi_harvester` to get the unique instance of this registry.')

        instance = cls.__instance__ = object.__new__(cls)
        return instance

    def __init__(self) -> None:
        """Initialize the Registry with thread locks and counters."""
        self.__count_lock: threading.Lock = threading.Lock()
        self.__count: int = 0

        self.__harvester_lock: Optional[HarvesterLock] = None
        self.__harvester: Optional[Harvester] = None

    def __enter__(self) -> HarvesterLock:
        """Enter the context manager, setting up the harvester."""
        return self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context manager, destroying the harvester."""
        return self.destroy()

    def setup(self) -> HarvesterLock:
        """
        Set up the harvester and register the instance.

        Returns:
            HarvesterLock: A proxy to the HarvesterLock instance.
        """
        with self.__count_lock:
            self.__count += 1
            if self.__count == 1:
                # Set up harvesters. Count lock is sufficient here.
                if self.__harvester is not None or self.__harvester_lock is not None:
                    if self.__harvester is not None:
                        logger.error('Harvester not deleted. Attempt setup either way.')
                    else:
                        logger.error('Harvester exit crashed. Attempt setup either way.')
                self.__harvester = Harvester()
                self.__harvester_lock = HarvesterLock(weakref.proxy(self.__harvester))
                self.__harvester.__enter__()

        return self.__harvester_lock

    def destroy(self, exc_type=None, exc_val=None, exc_tb=None) -> None:
        """Destroy the harvester instance and decrement the counter."""
        with self.__count_lock:
            if self.__count == 0:
                return
            self.__count -= 1
            if self.__count == 0:
                # Destroy the old instance
                with self.__harvester_lock:
                    self.__harvester_lock = None
                    try:
                        self.__harvester.__exit__(exc_type, exc_val, exc_tb)
                    finally:
                        # TODO: __exit__ is not supposed to throw here.
                        #       I created a fork of harvesters with a fix. Use that one.
                        #       If an exception is thrown here, we might not find any sensors.
                        self.__harvester = None

    def reset(self) -> None:
        """Reset the registry and reinitialize the harvester with old CTI files."""
        with self.__count_lock:
            # Back up the old owners
            old_owners: Dict[int, Set[str]] = self.__harvester_lock.owners

            # Destroy the old instance
            with self.__harvester_lock:
                self.__harvester_lock = None
                self.__harvester.__exit__(None, None, None)
                self.__harvester = None

            # Set up harvesters. Count lock is sufficient here.
            assert self.__harvester is None and self.__harvester_lock is None
            self.__harvester = Harvester()
            self.__harvester_lock = HarvesterLock(weakref.proxy(self.__harvester))
            self.__harvester.__enter__()

            # Add the old CTI files and update harvesters.
            for cti_files in old_owners.values():
                self.__harvester_lock.register(cti_files, False)
            self.__harvester_lock.update()


class MultiHarvester:
    """
    The unique MultiHarvester handle.

    Usage example for using the multi harvester in a with statement:

    >>> # Contains the path to the CTI file
    >>> cti_file: str

    >>> with MultiHarvester.registry as harvester_registry:
    >>>     slot = harvester_registry.register(cti_files={cti_file})
    >>>     with harvester_registry as unique_harvester_lock:
    >>>         unique_harvester_lock.harvester.device_info_list
    >>>     # ... acquire images here
    >>>     harvester_registry.deregister(slot)

    Usage example if the acquisition loop is not inside a single function:

    >>> harvester_registry = MultiHarvester.registry.setup()
    >>> slot = harvester_registry.register(cti_files={cti_file})
    >>> with harvester_registry as unique_harvester_lock:
    >>>     unique_harvester_lock.harvester.device_info_list
    >>> # ... acquire images here
    >>> harvester_registry.deregister(slot)
    >>> MultiHarvester.registry.destroy()
    >>> # harvester_registry cannot be accessed anymore
    """
    registry: Registry = Registry()


"""
with MultiHarvester.registry as harvester_registry:
    try:
        slot = harvester_registry.register(cti_files={
            '/opt/baumer-gapi-sdk-c/lib/libbgapi2_usb.cti'})
        with harvester_registry as unique_harvester_lock:
            dil = unique_harvester_lock.harvester.device_info_list
            print(dil)

        harvester_registry.deregister(slot)
    except Exception:
        traceback.print_exc()
"""
